/*
 * Driver to access Performance Monitoring Counters for
 * Mellanox BlueField SoC
 *
 * Copyright (c) 2018, Mellanox Technologies. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include <linux/acpi.h>
#include <linux/arm-smccc.h>
#include <linux/errno.h>
#include <linux/hwmon.h>
#include <linux/hwmon-sysfs.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/irqdomain.h>
#include <linux/io.h>
#include <linux/ioctl.h>
#include <linux/kernel.h>
#include <linux/kmod.h>
#include <linux/miscdevice.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/sysfs.h>
#include <linux/uaccess.h>
#include <linux/version.h>

#include <uapi/linux/psci.h>

#include "mlx-pmc.h"

#define DRIVER_VERSION		2.1
#define STRINGIFY(s)		#s

static struct pmc_context *pmc;

#define SIZE_64 0
#define SIZE_32 1

/* Calls an SMC to access a performance register */
static int secure_read(void *addr, int size, uint64_t *result)
{
	struct arm_smccc_res res;
	int status;
	uint32_t command;

	switch (size) {
	case SIZE_32:
		command = MLNX_READ_REG_32;
		break;
	case SIZE_64:
		command = MLNX_READ_REG_64;
		break;
	default:
		dev_err(pmc->hwmon_dev,
			"%s: invalid size: %d\n", __func__, size);
		return -1;
	}

	arm_smccc_smc(
		command,
		pmc->sreg_tbl_perf,
		(uintptr_t) addr,
		0, 0, 0, 0, 0, &res);

	status = res.a0;

	switch (status) {
	/*
	 * Note: PSCI_RET_NOT_SUPPORTED is used here to maintain compatibility
	 * with older kernels that do not have SMCCC_RET_NOT_SUPPORTED
	 */
	case PSCI_RET_NOT_SUPPORTED:
		dev_err(pmc->hwmon_dev,
			"%s: required SMC unsupported", __func__);
		return -1;
	case SMCCC_ACCESS_VIOLATION:
		dev_err(pmc->hwmon_dev,
			"%s: could not read register %p. Is it perf?",
			__func__,
			addr);
		return -1;
	default:
		*result = (uint64_t)res.a1;
		return 0;
	}
}

/* Read from a performance counter */
static int pmc_read(void *addr, int size, uint64_t *result)
{
	if (pmc->svc_sreg_support) {
		if (secure_read(addr, size, result))
			return -1;
		else
			return 0;
	} else {
		switch (size) {
		case SIZE_32:
			*result = (uint64_t)readl(addr);
			return 0;
		case SIZE_64:
			*result = readq(addr);
			return 0;
		default:
			dev_err(pmc->hwmon_dev,
				"%s: invalid size: %d\n", __func__, size);
			return -1;
		}
	}
}

/* Convenience function for 32-bit reads */
static int pmc_readl(uint32_t *result, void *addr)
{
	uint64_t read_out;
	int status;

	status = pmc_read(addr, SIZE_32, &read_out);
	if (status)
		return status;

	*result = (uint32_t)read_out;
	return 0;
}

/* Convenience function for 64-bit reads */
static int pmc_readq(uint64_t *result, void *addr)
{
	return pmc_read(addr, SIZE_64, result);
}

/* Calls an SMC to access a performance register */
static int secure_write(uint64_t value, void *addr, int size)
{
	struct arm_smccc_res res;
	int status;
	uint32_t command;

	switch (size) {
	case SIZE_32:
		command = MLNX_WRITE_REG_32;
		break;
	case SIZE_64:
		command = MLNX_WRITE_REG_64;
		break;
	default:
		dev_err(pmc->hwmon_dev,
			"%s: invalid size: %d\n", __func__, size);
		return -1;
	}

	arm_smccc_smc(
		command,
		pmc->sreg_tbl_perf,
		value,
		(uintptr_t) addr,
		0, 0, 0, 0, &res);

	status = res.a0;

	switch (status) {
	case PSCI_RET_NOT_SUPPORTED:
		dev_err(pmc->hwmon_dev,
			"%s: required SMC unsupported", __func__);
		return -1;
	case SMCCC_ACCESS_VIOLATION:
		dev_err(pmc->hwmon_dev,
			"%s: could not write register %p. Is it perf?",
			__func__,
			addr);
		return -1;
	default:
		return 0;
	}
}

/* Write to a performance counter */
static int pmc_write(uint64_t value, void *addr, int size)
{
	if (pmc->svc_sreg_support)
		return secure_write(value, addr, size);

	switch (size) {
	case SIZE_32:
		writel((uint32_t)value, addr);
		return 0;
	case SIZE_64:
		writeq(value, addr);
		return 0;
	default:
		dev_err(pmc->hwmon_dev,
			"%s: invalid size: %d\n", __func__, size);
		return -1;
	}
}

/* Convenience function for 32-bit writes */
static int pmc_writel(uint32_t value, void *addr)
{
	return pmc_write((uint64_t) value, addr, SIZE_32);
}

/* Convenience function for 64-bit writes */
static int pmc_writeq(uint64_t value, void *addr)
{
	return pmc_write(value, addr, SIZE_64);
}

/* Check if the register offset is within the mapped region for the block */
static bool valid_range(int blk_num, uint32_t offset)
{
	if (offset % 8 != 0)
		return false;  /* unaligned */
	if (offset >= 0 && offset + 8 <= pmc->block[blk_num].blk_size)
		return true;   /* inside the mapped PMC space */

	return false;
}

/* Get the block number using the name */
static int get_block_num(const char *name)
{
	int i = 0;

	for (i = 0; i < pmc->total_blocks; ++i)
		if (strcmp((char *)name, pmc->block_name[i]) == 0)
			return i;
	return -1;
}

/* Get the event list corresponding to a certain block */
struct perf_events *get_event_list(char *blk)
{
	struct perf_events *events = NULL;

	if (strstr(blk, "tile"))
		events = hnf_events;
	else if (strstr(blk, "triogen"))
		events = triogen_events;
	else if (strstr(blk, "trio"))
		events = trio_events;
	else if (strstr(blk, "mss"))
		events = mss_events;
	else if (strstr(blk, "ecc"))
		events = ecc_events;
	else if (strstr(blk, "pcie"))
		events = pcie_events;
	else if (strstr(blk, "l3cache"))
		events = l3cache_events;
	else
		return NULL;

	return events;
}

/* Get the event number given the name */
static int get_event_num(char *blk, char *evt)
{
	struct perf_events *events;
	int i = 0;

	events = get_event_list(blk);
	if (events == NULL)
		return -EINVAL;

	while (events[i].evt_name != NULL) {
		if (strcmp(evt, events[i].evt_name) == 0)
			return events[i].evt_num;
		++i;
	}

	return -1;
}

/* Get the event number given the name */
static char *get_event_name(char *blk, int evt)
{
	struct perf_events *events;
	int i = 0;

	events = get_event_list(blk);
	if (events == NULL)
		return NULL;

	while (events[i].evt_name != NULL) {
		if (evt == events[i].evt_num)
			return events[i].evt_name;
		++i;
	}

	return NULL;
}

/* Method to enable/disable/reset l3cache counters */
int config_l3_counters(int blk_num, bool enable, bool reset)
{
	L3C_PERF_CNT_CFG_t perfcnt_cfg;

	perfcnt_cfg.word = 0;
	if (enable)
		perfcnt_cfg.en = 1;
	if (reset)
		perfcnt_cfg.rst = 1;

	return pmc_writel(perfcnt_cfg.word, pmc->block[blk_num].mmio_base +
		L3C_PERF_CNT_CFG);
}


/* Method to handle l3cache counter programming */
int program_l3_counter(int blk_num, uint32_t cnt_num, uint32_t evt)
{
	L3C_PERF_CNT_SEL_t perfcnt_sel;
	L3C_PERF_CNT_SEL_1_t perfcnt_sel_1;
	void *pmcaddr;
	uint32_t *wordaddr;

	/* Disable all counters before programming them */
	if (config_l3_counters(blk_num, false, false))
		return -1;

	/* Select appropriate register information */
	switch (cnt_num) {
	case 0:
	case 1:
	case 2:
	case 3:
		pmcaddr = pmc->block[blk_num].mmio_base + L3C_PERF_CNT_SEL;
		wordaddr = &perfcnt_sel.word;
		break;
	case 4:
		pmcaddr = pmc->block[blk_num].mmio_base + L3C_PERF_CNT_SEL_1;
		wordaddr = &perfcnt_sel_1.word;
		break;
	default:
		return -1;
	}

	if (pmc_readl(wordaddr, pmcaddr))
		return -1;

	switch (cnt_num) {
	case 0:
		perfcnt_sel.cnt_0 = evt;
		break;
	case 1:
		perfcnt_sel.cnt_1 = evt;
		break;
	case 2:
		perfcnt_sel.cnt_2 = evt;
		break;
	case 3:
		perfcnt_sel.cnt_3 = evt;
		break;
	case 4:
		perfcnt_sel_1.cnt_4 = evt;
		break;
	default:
		return -1;
	}

	if (pmc_writel(*wordaddr, pmcaddr))
		return -1;

	return 0;
}

/* Method to program a counter to monitor an event */
int program_counter(int blk_num, uint32_t cnt_num, uint32_t evt, bool is_l3)
{
	GEN_PERFCTL_t perfctl;
	GEN_PERFEVT_t perfevt;
	GEN_PERFMON_CONFIG_t perfmon_cfg;

	if (cnt_num >= pmc->block[blk_num].counters)
		return -1;

	if (is_l3)
		return program_l3_counter(blk_num, cnt_num, evt);

	/* Configure the counter */
	perfctl.word = 0;
	perfctl.en0 = 1;
	perfctl.eb0 = 0;
	perfctl.etrig0 = 1;
	perfctl.ad0 = 0;
	perfctl.accm0 = 0;
	perfctl.ms0 = 0;
	perfctl.fm0 = 0;

	perfmon_cfg.word = 0;
	perfmon_cfg.wdata = perfctl.word;
	perfmon_cfg.addr = PERFCTL;
	perfmon_cfg.strobe = 1;
	perfmon_cfg.wr_rd_b = 1;

	if (pmc_writeq(perfmon_cfg.word,
				pmc->block[blk_num].mmio_base + cnt_num * 8))
		return -1;

	/* Select the event */
	perfevt.word = 0;
	perfevt.evtsel = evt;

	perfmon_cfg.word = 0;
	perfmon_cfg.wdata = perfevt.word;
	perfmon_cfg.addr = PERFEVT;
	perfmon_cfg.strobe = 1;
	perfmon_cfg.wr_rd_b = 1;

	if (pmc_writeq(perfmon_cfg.word,
				pmc->block[blk_num].mmio_base + cnt_num * 8))
		return -1;

	/* Clear the accumulator */
	perfmon_cfg.word = 0;
	perfmon_cfg.addr = PERFACC0;
	perfmon_cfg.strobe = 1;
	perfmon_cfg.wr_rd_b = 1;

	if (pmc_writeq(perfmon_cfg.word,
				pmc->block[blk_num].mmio_base + cnt_num * 8))
		return -1;

	return 0;
}

/* Method to handle l3 counter reads */
int read_l3_counter(int blk_num, uint32_t cnt_num, uint64_t *result)
{
	L3C_PERF_CNT_LOW_t perfcnt_low;
	L3C_PERF_CNT_HIGH_t perfcnt_high;
	uint64_t value;
	int status = 0;

	status = pmc_readl(&perfcnt_low.word,
		pmc->block[blk_num].mmio_base +
		L3C_PERF_CNT_LOW + cnt_num * 4);

	if (status)
		return -1;

	status = pmc_readl(&perfcnt_high.word,
		pmc->block[blk_num].mmio_base +
		L3C_PERF_CNT_HIGH + cnt_num * 4);

	if (status)
		return -1;

	value = perfcnt_high.val;
	value = value << 32;
	value |= perfcnt_low.val;

	*result = value;
	return 0;
}

/* Method to read the counter value */
int read_counter(int blk_num, uint32_t cnt_num, bool is_l3,
		uint64_t *result)
{
	GEN_PERFMON_CONFIG_t perfmon_cfg;
	uint32_t perfcfg_offset, perfval_offset;
	int status;

	if (cnt_num >= pmc->block[blk_num].counters)
		return -1;

	if (is_l3)
		return read_l3_counter(blk_num, cnt_num, result);

	perfcfg_offset = cnt_num * 8;
	perfval_offset = perfcfg_offset + pmc->block[blk_num].counters * 8;

	/* Set counter in "read" mode */
	perfmon_cfg.word = 0;
	perfmon_cfg.addr = PERFACC0;
	perfmon_cfg.strobe = 1;
	perfmon_cfg.wr_rd_b = 0;

	status = pmc_writeq(perfmon_cfg.word, pmc->block[blk_num].mmio_base +
		perfcfg_offset);

	if (status)
		return -1;

	/* Get the counter value */
	return pmc_readq(result,
		pmc->block[blk_num].mmio_base + perfval_offset);
}

int read_l3_event(int blk_num, uint32_t cnt_num, uint64_t *result)
{
	L3C_PERF_CNT_SEL_t perfcnt_sel;
	L3C_PERF_CNT_SEL_1_t perfcnt_sel_1;
	uint64_t evt;
	void *pmcaddr;
	uint32_t *wordaddr;

	/* Select appropriate register information */
	switch (cnt_num) {
	case 0:
	case 1:
	case 2:
	case 3:
		pmcaddr = pmc->block[blk_num].mmio_base + L3C_PERF_CNT_SEL;
		wordaddr = &perfcnt_sel.word;
		break;
	case 4:
		pmcaddr = pmc->block[blk_num].mmio_base + L3C_PERF_CNT_SEL_1;
		wordaddr = &perfcnt_sel_1.word;
		break;
	default:
		return -1;
	}

	if (pmc_readl(wordaddr, pmcaddr))
		return -1;

	/* Read from appropriate register field for the counter */
	switch (cnt_num) {
	case 0:
		evt = perfcnt_sel.cnt_0;
		break;
	case 1:
		evt = perfcnt_sel.cnt_1;
		break;
	case 2:
		evt = perfcnt_sel.cnt_2;
		break;
	case 3:
		evt = perfcnt_sel.cnt_3;
		break;
	case 4:
		evt = perfcnt_sel_1.cnt_4;
		break;
	default:
		return -1;
	}

	*result = evt;
	return 0;
}

/* Method to find the event currently being monitored by a counter */
int read_event(int blk_num, uint32_t cnt_num, bool is_l3,
		uint64_t *result)
{
	GEN_PERFMON_CONFIG_t perfmon_cfg;
	GEN_PERFEVT_t perfevt;
	GEN_PERFCTL_t perfctl;
	uint32_t perfcfg_offset, perfval_offset;
	uint64_t tmpval;

	if (cnt_num >= pmc->block[blk_num].counters)
		return -1;

	if (is_l3)
		return read_l3_event(blk_num, cnt_num, result);

	perfcfg_offset = cnt_num * 8;
	perfval_offset = perfcfg_offset + pmc->block[blk_num].counters * 8;

	/* Set counter in "read" mode */
	perfmon_cfg.word = 0;
	perfmon_cfg.addr = PERFCTL;
	perfmon_cfg.strobe = 1;
	perfmon_cfg.wr_rd_b = 0;

	if (pmc_writeq(perfmon_cfg.word, pmc->block[blk_num].mmio_base +
		perfcfg_offset))
		return -1;

	/* Check if the counter is enabled */

	if (pmc_readq(&tmpval,
		pmc->block[blk_num].mmio_base + perfval_offset))
		return -1;
	perfctl.word = (uint32_t)tmpval;

	if (perfctl.en0 == 0)
		return 0xff;

	/* Set counter in "read" mode */
	perfmon_cfg.word = 0;
	perfmon_cfg.addr = PERFEVT;
	perfmon_cfg.strobe = 1;
	perfmon_cfg.wr_rd_b = 0;


	if (pmc_writeq(perfmon_cfg.word, pmc->block[blk_num].mmio_base +
		perfcfg_offset))
		return -1;

	/* Get the event number */
	if (pmc_readq(&tmpval,
		pmc->block[blk_num].mmio_base + perfval_offset))
		return -1;
	perfevt.word = (uint32_t)tmpval;

	*result = perfevt.evtsel;

	return 0;
}

/* Method to read a register */
int read_reg(int blk_num, uint32_t offset, uint64_t *result)
{
	uint32_t ecc_out;

	if (strstr(pmc->block_name[blk_num], "ecc")) {
		if (pmc_readl(&ecc_out,
			pmc->block[blk_num].mmio_base + offset))
			return -1;

		*result = (uint64_t) ecc_out;
		return 0;
	}

	if (valid_range(blk_num, offset))
		return pmc_readq(result,
			pmc->block[blk_num].mmio_base + offset);

	return -1;
}

/* Method to write to a register */
int write_reg(int blk_num, uint32_t offset, uint64_t data)
{
	if (strstr(pmc->block_name[blk_num], "ecc")) {
		return pmc_writel((uint32_t)data,
				pmc->block[blk_num].mmio_base + offset);
	}

	if (valid_range(blk_num, offset))
		return pmc_writeq(data,
			pmc->block[blk_num].mmio_base + offset);

	return -1;
}

/* Show function for "counter" sysfs files */
static ssize_t counter_read(struct kobject *ko, struct kobj_attribute *attr,
				char *buf)
{
	int blk_num, cnt_num, offset, err;
	uint64_t value;
	bool is_l3 = false;

	blk_num = get_block_num(ko->name);
	if (blk_num == -1)
		return -EINVAL;

	if (strstr(ko->name, "l3cache"))
		is_l3 = true;

	if (pmc->block[blk_num].type == PERFTYPE_COUNTER) {
		err = sscanf(attr->attr.name, "counter%d", &cnt_num);
		if (err == -1)
			return -EINVAL;
		if (read_counter(blk_num, cnt_num, is_l3, &value))
			return -EINVAL;
	} else  if (pmc->block[blk_num].type == PERFTYPE_REGISTER) {
		offset = get_event_num((char *)ko->name,
			(char *)attr->attr.name);
		if (offset == -1)
			return -EINVAL;
		if (read_reg(blk_num, offset, &value))
			return -EINVAL;
	} else
		return -EINVAL;

	return sprintf(buf, "0x%llx\n", value);
}

/* Store function for "counter" sysfs files */
static ssize_t counter_clear(struct kobject *ko, struct kobj_attribute *attr,
				const char *buf, size_t count)
{
	int blk_num, cnt_num, offset, err, data;
	uint64_t evt_num;
	bool is_l3 = false;

	blk_num = get_block_num(ko->name);
	if (blk_num == -1)
		return -EINVAL;

	err = sscanf(buf, "%x\n", &data);
	if (err == -1)
		return -EINVAL;

	/* Allow non-zero writes only to the ecc regs */
	if (!(strstr(ko->name, "ecc")) && (data != 0))
		return -EINVAL;

	if (strstr(ko->name, "l3cache"))
		return -EINVAL;

	if (pmc->block[blk_num].type == PERFTYPE_COUNTER) {
		err = sscanf(attr->attr.name, "counter%d", &cnt_num);
		if (err == -1)
			return -EINVAL;
		err = read_event(blk_num, cnt_num, is_l3, &evt_num);
		if (err == -1)
			return -EINVAL;
		err = program_counter(blk_num, cnt_num, evt_num, is_l3);
		if (err == -1)
			return -EINVAL;
	} else if (pmc->block[blk_num].type == PERFTYPE_REGISTER) {
		offset = get_event_num((char *)ko->name,
			(char *)attr->attr.name);
		if (offset == -1)
			return -EINVAL;
		err = write_reg(blk_num, offset, data);
		if (err == -1)
			return -EINVAL;
	} else
		return -EINVAL;

	return count;
}

/* Show function for "event" sysfs files */
static ssize_t event_find(struct kobject *ko, struct kobj_attribute *attr,
				char *buf)
{
	int blk_num, cnt_num, err;
	uint64_t evt_num;
	char *evt_name;
	bool is_l3 = false;

	blk_num = get_block_num(ko->name);
	if (blk_num == -1)
		return -EINVAL;

	if (strstr(ko->name, "l3cache"))
		is_l3 = true;

	err = sscanf(attr->attr.name, "event%d", &cnt_num);
	if (err == -1)
		return -EINVAL;

	err = read_event(blk_num, cnt_num, is_l3, &evt_num);
	if (err == -1)
		return -EINVAL;

	evt_name = get_event_name((char *)ko->name, evt_num);

	return sprintf(buf, "0x%llx: %s\n", evt_num, evt_name);
}

/* Store function for "event" sysfs files */
static ssize_t event_set(struct kobject *ko, struct kobj_attribute *attr,
				const char *buf, size_t count)
{
	int blk_num, cnt_num, evt_num, err;
	bool is_l3 = false;

	if (isalpha(buf[0])) {
		evt_num = get_event_num((char *)ko->name, (char *)buf);
		if (evt_num == -1)
			return -EINVAL;
	} else {
		err = sscanf(buf, "%x\n", &evt_num);
		if (err == -1)
			return -EINVAL;
	}

	blk_num = get_block_num(ko->name);
	if (blk_num == -1)
		return -EINVAL;

	err = sscanf(attr->attr.name, "event%d", &cnt_num);
	if (err == -1)
		return -EINVAL;

	if (strstr(ko->name, "l3cache"))
		is_l3 = true;

	err = program_counter(blk_num, cnt_num, evt_num, is_l3);
	if (err == -1)
		return -EINVAL;

	return count;
}

/* Show function for "event_list" sysfs files */
static ssize_t print_event_list(struct kobject *ko,
			struct kobj_attribute *attr, char *buf)
{
	struct perf_events *events;
	int i = 0, size = 0, ret = 0;
	char e_info[100];

	events = get_event_list((char *)ko->name);
	if (events == NULL)
		return -EINVAL;

	buf[0] = '\0';
	while (events[i].evt_name != NULL) {
		size += sprintf(e_info, "%x: %s\n", events[i].evt_num,
			events[i].evt_name);
		if (size > PAGE_SIZE)
			break;
		strcat(buf, e_info);
		ret = size;
		++i;
	}

	return ret;
}

/* Show function for "enable" sysfs files - only for l3cache */
static ssize_t show_state_counters(struct kobject *ko,
			struct kobj_attribute *attr, char *buf)
{
	L3C_PERF_CNT_CFG_t perfcnt_cfg;
	int blk_num, value;

	blk_num = get_block_num(ko->name);
	if (blk_num == -1)
		return -EINVAL;

	if (pmc_readl(&perfcnt_cfg.word, pmc->block[blk_num].mmio_base +
		L3C_PERF_CNT_CFG))
		return -EINVAL;

	value = perfcnt_cfg.en;

	return sprintf(buf, "%d\n", value);
}

/* Store function for "enable" sysfs files - only for l3cache */
static ssize_t enable_counters(struct kobject *ko, struct kobj_attribute *attr,
				const char *buf, size_t count)
{
	int err, en, blk_num;

	blk_num = get_block_num(ko->name);
	if (blk_num == -1)
		return -EINVAL;

	err = sscanf(buf, "%x\n", &en);
	if (en == 0) {
		if (config_l3_counters(blk_num, false, false))
			return -EINVAL;
	} else if (en == 1) {
		if (config_l3_counters(blk_num, false, true))
			return -EINVAL;
		if (config_l3_counters(blk_num, true, false))
			return -EINVAL;
	} else
		return -EINVAL;

	return count;
}

/* Helper to create the bfperf sysfs sub-directories and files */
int bfperf_create_sysfs(struct device *dev, struct kobject *ko, int blk_num)
{
	int err = 0, j = 0;

	pmc->block[blk_num].block_dir =
		kobject_create_and_add(pmc->block_name[blk_num], ko);
	if (pmc->block[blk_num].block_dir == NULL) {
		dev_err(dev,
			"PMC: Error creating subdirectories\n");
		return -EFAULT;
	}

	if (pmc->block[blk_num].type == PERFTYPE_COUNTER) {
		pmc->block[blk_num].attr_event_list.attr.mode = 0444;
		pmc->block[blk_num].attr_event_list.show = print_event_list;
		pmc->block[blk_num].attr_event_list.attr.name =
			kzalloc(20, GFP_KERNEL);
		snprintf((char *)pmc->block[blk_num].attr_event_list.attr.name,
			20, "event_list");

		err = sysfs_create_file_ns(pmc->block[blk_num].block_dir,
			&pmc->block[blk_num].attr_event_list.attr, NULL);
		if (err < 0) {
			dev_err(dev,
			"PMC: Error creating sysfs entries\n");
			return err;
		}

		if (strstr(pmc->block_name[blk_num], "l3cache")) {
			pmc->block[blk_num].attr_enable.attr.mode =
				0644;
			pmc->block[blk_num].attr_enable.show =
				show_state_counters;
			pmc->block[blk_num].attr_enable.store =
				enable_counters;
			pmc->block[blk_num].attr_enable.attr.name =
				kzalloc(20, GFP_KERNEL);
			snprintf((char *)
				pmc->block[blk_num].attr_enable.attr.name,
				20, "enable");

			err = sysfs_create_file_ns(
				pmc->block[blk_num].block_dir,
				&pmc->block[blk_num].attr_enable.attr,
				NULL);
			if (err < 0) {
				dev_err(dev,
				"PMC: Error creating sysfs entries\n");
				return err;
			}

		}

		pmc->block[blk_num].attr_counter =
			kcalloc(pmc->block[blk_num].counters,
				sizeof(struct kobj_attribute), GFP_KERNEL);
		if (!pmc->block[blk_num].attr_counter)
			return -ENOMEM;
		pmc->block[blk_num].attr_event =
			kcalloc(pmc->block[blk_num].counters,
				sizeof(struct kobj_attribute), GFP_KERNEL);
		if (!pmc->block[blk_num].attr_event)
			return -ENOMEM;
		pmc->block[blk_num].sysfs_event_cnt =
			pmc->block[blk_num].counters;

		for (j = 0; j < pmc->block[blk_num].counters; ++j) {
			pmc->block[blk_num].attr_counter[j].attr.mode = 0644;
			pmc->block[blk_num].attr_counter[j].show =
				counter_read;
			pmc->block[blk_num].attr_counter[j].store =
				counter_clear;
			pmc->block[blk_num].attr_counter[j].attr.name =
				kzalloc(20, GFP_KERNEL);
			snprintf((char *)
				pmc->block[blk_num].attr_counter[j].attr.name,
				20, "counter%d", j);

			err = sysfs_create_file_ns(
				pmc->block[blk_num].block_dir,
				&pmc->block[blk_num].attr_counter[j].attr,
				NULL);
			if (err < 0) {
				dev_err(dev,
				"PMC: Error creating sysfs entries\n");
				return err;
			}

			pmc->block[blk_num].attr_event[j].attr.mode = 0644;
			pmc->block[blk_num].attr_event[j].show = event_find;
			pmc->block[blk_num].attr_event[j].store = event_set;
			pmc->block[blk_num].attr_event[j].attr.name =
				kzalloc(20, GFP_KERNEL);
			snprintf((char *)
				pmc->block[blk_num].attr_event[j].attr.name,
				20, "event%d", j);

			err = sysfs_create_file_ns(
				pmc->block[blk_num].block_dir,
				&pmc->block[blk_num].attr_event[j].attr,
				NULL);
			if (err < 0) {
				dev_err(dev,
				"PMC: Error creating sysfs entries\n");
				return err;
			}
		}
	} else if (pmc->block[blk_num].type == PERFTYPE_REGISTER) {
		struct perf_events *events;

		events = get_event_list((char *)pmc->block_name[blk_num]);
		if (events == NULL)
			return -EINVAL;

		while (events[j].evt_name != NULL)
			++j;

		pmc->block[blk_num].sysfs_event_cnt = j;
		pmc->block[blk_num].attr_event =
			kcalloc(j, sizeof(struct kobj_attribute), GFP_KERNEL);
		if (!pmc->block[blk_num].attr_event)
			return -ENOMEM;

		while (j > 0) {
			--j;
			pmc->block[blk_num].attr_event[j].attr.mode = 0644;
			pmc->block[blk_num].attr_event[j].show = counter_read;
			pmc->block[blk_num].attr_event[j].store = counter_clear;
			pmc->block[blk_num].attr_event[j].attr.name =
				kzalloc(30, GFP_KERNEL);
			strcpy((char *)
				pmc->block[blk_num].attr_event[j].attr.name,
				events[j].evt_name);

			err = sysfs_create_file_ns(
				pmc->block[blk_num].block_dir,
				&pmc->block[blk_num].attr_event[j].attr,
				NULL);
			if (err < 0) {
				dev_err(dev,
				"PMC: Error creating sysfs entries\n");
				return err;
			}
		}
	} else
		err = -EINVAL;

	return err;
}

void pmc_delete(void)
{
	hwmon_device_unregister(pmc->hwmon_dev);
	kfree(pmc);
}

static int pmc_probe(struct platform_device *pdev)
{
	struct device *dev = &pdev->dev;
	int i, version, err = 0, ret = 0;
	uint64_t info[4];
	struct arm_smccc_res res;

	/*
	 * Ensure we have the UUID we expect for the Mellanox service.
	 */
	arm_smccc_smc(MLNX_SIP_SVC_UID, 0, 0, 0, 0, 0, 0, 0, &res);
	if (res.a0 != 0x89c036b4 || res.a1 != 0x11e6e7d7 ||
		res.a2 != 0x1a009787 || res.a3 != 0xc4bf00ca)
		return -ENODEV;

	pmc = kzalloc(sizeof(struct pmc_context), GFP_KERNEL);
	if (!pmc)
		return -ENOMEM;

	platform_set_drvdata(pdev, pmc);
	pmc->pdev = pdev;

	pmc->hwmon_dev = hwmon_device_register_with_info(dev, "bfperf", pmc,
		NULL, NULL);
	pmc->ko = &pmc->hwmon_dev->kobj;

	/*
	 * ACPI indicates whether we use SMCs to access registers or not.
	 * If sreg_tbl_perf is not present, just assume we're not using SMCs.
	 */
	if (device_property_read_u32(dev,
				"sec_reg_block", &pmc->sreg_tbl_perf)) {
		pmc->svc_sreg_support = false;
	} else {
		/*
		 * Check service version to see if we actually do support the
		 * needed SMCs. If we have the calls we need, mark support for
		 * them in the pmc struct.
		 */
		arm_smccc_smc(MLNX_SIP_SVC_VERSION, 0, 0, 0, 0, 0, 0, 0, &res);
		if (res.a0 == MLNX_PMC_SVC_REQ_MAJOR &&
			res.a1 >= MLNX_PMC_SVC_MIN_MINOR)
			pmc->svc_sreg_support = true;
		else {
			dev_err(dev, "Required SMCs are not supported.\n");

			err = -EINVAL;
			goto error;
		}
	}

	if (pmc->ko == NULL) {
		dev_err(dev, "Sysfs creation failed\n");
		err = -EFAULT;
		goto error;
	}

	if (device_property_read_u32(dev, "version", &version)) {
		dev_err(dev, "Version Info not found\n");
		err = -EINVAL;
		goto error;
	}

	if (version != (int)DRIVER_VERSION) {
		dev_err(dev, "Version Mismatch. Expected %d Returned %d\n",
			(int)DRIVER_VERSION, version);
		err = -EINVAL;
		goto error;
	}

	if (device_property_read_u32(dev, "block_num", &pmc->total_blocks)) {
		dev_err(dev, "Number of performance blocks undefined\n");
		err = -EINVAL;
		goto error;
	}

	ret = device_property_read_string_array(dev, "block_name",
		pmc->block_name, pmc->total_blocks);
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 13, 0)
	if (ret) {
		dev_err(dev, "Reading block names failed\n");
		err = ret;
		goto error;
	}
#else
	if (ret != pmc->total_blocks) {
		dev_err(dev,
			"Block count mismatch. Expected %d Returned %d\n",
			pmc->total_blocks, ret);
		err = -EFAULT;
		goto error;
	}
#endif

	if (device_property_read_u32(dev, "tile_num", &pmc->tile_count)) {
		dev_err(dev, "Number of tiles undefined\n");
		err = -EINVAL;
		goto error;
	}

	/* Map the Performance Counters from the varios blocks */
	for (i = 0; i < pmc->total_blocks; ++i) {
		if (strstr(pmc->block_name[i], "tile")) {
			int tile_num;

			ret = sscanf(pmc->block_name[i], "tile%d", &tile_num);
			if (ret == -1) {
				err = -EINVAL;
				goto error;
			}
			if (tile_num >= pmc->tile_count)
				continue;
		}
		err = device_property_read_u64_array(dev, pmc->block_name[i],
			info, 4);
		if (err) {
			dev_err(dev, "Failed to find %s block info\n",
				pmc->block_name[i]);
			goto error;
		}

		/*
		 * Do not remap if the proper SMC calls are supported, since the
		 * SMC calls expect physical addresses.
		 */
		if (pmc->svc_sreg_support)
			pmc->block[i].mmio_base = (void *)info[0];
		else
			pmc->block[i].mmio_base = ioremap(info[0], info[1]);

		pmc->block[i].blk_size = info[1];
		pmc->block[i].counters = info[2];
		pmc->block[i].type = info[3];

		if (IS_ERR(pmc->block[i].mmio_base)) {
			dev_err(dev, "%s: ioremap failed base %llx err %p\n",
				__func__, info[0], pmc->block[i].mmio_base);
			err = PTR_ERR(pmc->block[i].mmio_base);
			goto error;
		}

		err = bfperf_create_sysfs(dev, pmc->ko, i);
	}

	dev_info(&pdev->dev, "v%d probed\n", (int)DRIVER_VERSION);
	return 0;

error:
	pmc_delete();
	return err;
}

static int pmc_remove(struct platform_device *pdev)
{
	struct pmc_context *pmc = platform_get_drvdata(pdev);
	int i, j, err;

	for (i = 0; i < pmc->total_blocks; ++i) {
		if (strstr(pmc->block_name[i], "tile")) {
			int tile_num;

			err = sscanf(pmc->block_name[i], "tile%d", &tile_num);
			if (err == -1)
				return -EINVAL;
			if (tile_num >= pmc->tile_count)
				continue;
		}
		kfree(pmc->block[i].attr_event_list.attr.name);
		if (pmc->block[i].type == PERFTYPE_COUNTER) {
			for (j = 0; j < pmc->block[i].counters; ++j) {
				kfree(pmc->block[i].attr_counter[j].attr.name);
				kfree(pmc->block[i].attr_event[j].attr.name);
			}
		} else if (pmc->block[i].type == PERFTYPE_REGISTER) {
			for (j = 0; j < pmc->block[i].sysfs_event_cnt; ++j)
				kfree(pmc->block[i].attr_event[j].attr.name);
		}

		/* Unmap if SMCs weren't used for access */
		if (pmc->block[i].mmio_base && !(pmc->svc_sreg_support))
			iounmap(pmc->block[i].mmio_base);

		kobject_put(pmc->block[i].block_dir);
		kfree(pmc->block[i].attr_event);
		kfree(pmc->block[i].attr_counter);
	}

	pmc_delete();

	return 0;
}

static const struct acpi_device_id pmc_acpi_ids[] = {
	{"MLNXBFD0", 0},
	{},
};

MODULE_DEVICE_TABLE(acpi, pmc_acpi_ids);
static struct platform_driver pmc_driver = {
	.driver = {
		.name = "mlx-pmc",
		.acpi_match_table = ACPI_PTR(pmc_acpi_ids),
	},
	.probe = pmc_probe,
	.remove = pmc_remove,
};

module_platform_driver(pmc_driver);

MODULE_AUTHOR("Mellanox Technologies");
MODULE_DESCRIPTION("Mellanox PMC driver");
MODULE_LICENSE("GPL");
MODULE_VERSION(STRINGIFY(DRIVER_VERSION));

