/*
 * Copyright (c) 2018, Mellanox Technologies. All rights reserved.
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */
#ifndef __MLX_PMC_H__
#define __MLX_PMC_H__

#include "mlx-perf-events.h"

#define MAX_NUM_BLOCKS		30

#define MLNX_WRITE_REG_32		(0x82000009)
#define MLNX_READ_REG_32		(0x8200000A)
#define MLNX_WRITE_REG_64		(0x8200000B)
#define MLNX_READ_REG_64		(0x8200000C)
#define MLNX_SIP_SVC_UID		(0x8200ff01)
#define MLNX_SIP_SVC_VERSION		(0x8200ff03)

#define SMCCC_INVALID_PARAMETERS	(-2)
#define SMCCC_OUT_OF_RANGE		(-3)
#define SMCCC_ACCESS_VIOLATION		(-4)

#define MLNX_PMC_SVC_REQ_MAJOR 0
#define MLNX_PMC_SVC_MIN_MINOR 3

struct block_info {
	/* The VA at which the PMC block is mapped. */
	void *mmio_base;

	/* Size of each mapped region */
	size_t blk_size;

	/* Number of counters in the block */
	size_t counters;

	/* Type of counters in the block */
	int type;

	/* Kobjects to create sub-directories */
	struct kobject *block_dir;

	/* Attributes for "counter" sysfs files */
	struct kobj_attribute *attr_counter;

	/* Attributes for "event" sysfs files */
	struct kobj_attribute *attr_event;

	/* Attributes for "event_list" sysfs files */
	struct kobj_attribute attr_event_list;

	/* Attributes for "enable" sysfs files */
	struct kobj_attribute attr_enable;

	/* Number of sysfs event files in the block */
	int sysfs_event_cnt;
};

struct pmc_context {
	/* The kernel structure representing the device. */
	struct platform_device *pdev;

	/* Total number of blocks */
	uint32_t total_blocks;

	/* Number of tiles in the system */
	uint32_t tile_count;

	/* Hwmon device for bfperf */
	struct device *hwmon_dev;

	/* Kobject for bfperf */
	struct kobject *ko;

	/* Block name */
	const char *block_name[MAX_NUM_BLOCKS];

	/* Block info */
	struct block_info block[MAX_NUM_BLOCKS];

	/* Determines whether SMCs are used to access performance registers */
	bool svc_sreg_support;

	/* Secure register access table number */
	uint32_t sreg_tbl_perf;
};

#define PERFTYPE_COUNTER	1
#define PERFTYPE_REGISTER	0

#define PERFCTL 0
#define PERFEVT 1
#define PERFVALEXT 2
#define PERFACC0 4
#define PERFACC1 5
#define PERFMVAL0 6
#define PERFMVAL1 7

typedef union {
	struct {
		/* Write(1) or read(0) control */
		uint64_t wr_rd_b:1;
		/* access strobe (pulse) */
		uint64_t strobe:1;
		/* Configuration access address offset */
		uint64_t addr:3;
		/* Configuration write data */
		uint64_t wdata:56;
		/* Reserved. */
		uint64_t __reserved:3;
	};

	uint64_t word;
} GEN_PERFMON_CONFIG_t;

typedef union {
	struct {
		/* Selected generic performance register's value */
		uint64_t val:64;
	};

	uint64_t word;
} GEN_PERFMVAL_t;

typedef union {
	struct {
		uint32_t fm1:3;
		uint32_t reserved_4:1;
		uint32_t ms1:2;
		uint32_t reserved_3:2;
		uint32_t accm1:3;
		uint32_t ad1:1;
		uint32_t etrig1:2;
		uint32_t eb1:1;
		uint32_t en1:1;
		uint32_t fm0:3;
		uint32_t reserved_2:1;
		uint32_t ms0:2;
		uint32_t reserved_1:2;
		uint32_t accm0:3;
		uint32_t ad0:1;
		uint32_t etrig0:2;
		uint32_t eb0:1;
		uint32_t en0:1;
	};
	uint32_t word;
} GEN_PERFCTL_t;

typedef union {
	struct {
		uint32_t reserved_1:16;
		uint32_t pvalsel:4;
		uint32_t modsel:4;
		uint32_t evtsel:8;
	};
	uint32_t word;
} GEN_PERFEVT_t;

#define L3C_PERF_CNT_CFG 0x0
#define L3C_PERF_CNT_CFG_1 0x4
#define L3C_PERF_CNT_CFG_2 0x8
#define L3C_PERF_CNT_SEL 0x10
#define L3C_PERF_CNT_SEL_1 0x14
#define L3C_PERF_CNT_LOW 0x40
#define L3C_PERF_CNT_HIGH 0x60

typedef union {
	struct {
		/* 1 - enable performance counters. */
		uint32_t en:1;
		/* 1 - reset all performance counters. */
		uint32_t rst:1;
		/* Reserved. */
		uint32_t __reserved_0:6;
		/* source id select, for relevant counters */
		uint32_t srcid_sel:7;
		/* Reserved. */
		uint32_t __reserved_1:1;
		/*
		 * source id mask, wildcard selecting with srcid_sel
		 * 1 - compare (matching srcid_sel bit is valid).
		 * 0 - ignore matching srcid_sel bit.
		 */
		uint32_t srcid_mask:7;
		/* Reserved. */
		uint32_t __reserved_2:1;
		/* profile select, for relevant counters */
		uint32_t prf_sel:4;
		/*
		 * profile mask, wildcard selecting with prf_sel
		 * 1 - compare (matching prf_sel bit is valid).
		 * 0 - ignore matching prf_sel bit.
		 */
		uint32_t prf_mask:4;
	};
	uint32_t word;
} L3C_PERF_CNT_CFG_t;

typedef union {
	struct {
		/* set select, for relevant counters */
		uint32_t set_sel:11;
		/* Reserved. */
		uint32_t __reserved_0:1;
		/*
		 * set mask, wildcard selecting with set_sel
		 * 1 - compare (matching set_sel bit is valid).
		 * 0 - ignore matching set_sel bit.
		 */
		uint32_t set_mask:11;
		/* Reserved. */
		uint32_t __reserved_1:1;
		/*
		 * count relevant counters when the usage is greater/equal to
		 * this threshold
		 */
		uint32_t emem_usage_th:7;
		/* Reserved. */
		uint32_t __reserved_2:1;
	};
	uint32_t word;
} L3C_PERF_CNT_CFG_1_t;

typedef union {
	struct {
		/*
		 * profile select in bits [3:0], for relevant counters
		 * [7:4] - must be "0".
		 */
		uint32_t strm_sel:8;
		/*
		 * stream mask, wildcard selecting with strm_sel
		 * 1 - compare (matching strm_sel bit is valid).
		 * 0 - ignore matching strm_sel bit.
		 */
		uint32_t strm_mask:8;
		/* Reserved. */
		uint32_t __reserved:16;
	};
	uint32_t word;
} L3C_PERF_CNT_CFG_2_t;

typedef union {
	struct {
		/* counter select */
		uint32_t cnt_0:6;
		/* Reserved. */
		uint32_t __reserved_0:2;
		/* counter select */
		uint32_t cnt_1:6;
		/* Reserved. */
		uint32_t __reserved_1:2;
		/* counter select */
		uint32_t cnt_2:6;
		/* Reserved. */
		uint32_t __reserved_2:2;
		/* counter select */
		uint32_t cnt_3:6;
		/* Reserved. */
		uint32_t __reserved_3:2;
	};
	uint32_t word;
} L3C_PERF_CNT_SEL_t;

typedef union {
	struct {
		/* counter select */
		uint32_t cnt_4:6;
		/* Reserved. */
		uint32_t __reserved:26;
	};
	uint32_t word;
} L3C_PERF_CNT_SEL_1_t;

typedef union {
	struct {
		/* counter */
		uint32_t val:32;
	};
	uint32_t word;
} L3C_PERF_CNT_LOW_t;

typedef union {
	struct {
		/* counter */
		uint32_t val:24;
		/* Reserved. */
		uint32_t __reserved:8;
	};
	uint32_t word;
} L3C_PERF_CNT_HIGH_t;

#endif				/* __MLX_PMC_H__ */
